/*
  Copyright ©2013 The Regents of the University of California
  (Regents). All Rights Reserved. Permission to use, copy, modify, and
  distribute this software and its documentation for educational,
  research, and not-for-profit purposes, without fee and without a
  signed licensing agreement, is hereby granted, provided that the
  above copyright notice, this paragraph and the following two
  paragraphs appear in all copies, modifications, and
  distributions. Contact The Office of Technology Licensing, UC
  Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620,
  (510) 643-7201, for commercial licensing opportunities.

  IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
  DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGE.

  REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING
  DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
  IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/

#include "display.hpp"

#include "bvh.hpp"
#include "geometry.hpp"
#include "io.hpp"
#include "timer.hpp"
#include "opengl.hpp"
#include "timer.hpp"
#include "util.hpp"
#include <sstream>

using namespace std;

static vector<Mesh*> &meshes = sim.cloth_meshes;
string obj2png_filename;

extern int frame;
extern Timer fps;

struct View {
    double lat, lon;
    Vec2 offset;
    double scale;
    View (): lat(0), lon(0), offset(0), scale(0.5) {}
};

enum Pane {MaterialPane, PlasticPane, WorldPane};

bool pane_enabled[3] = {true, true, true};
int subwindows[3];
View views[3];

int get_pane () {return find(glutGetWindow(), subwindows, 3);}

void reshape (int w, int h) {
    int npanes = 0;
    for (int i = 0; i < 3; i++)
        if (pane_enabled[i])
            npanes++;
    int j = 0;
    for (int i = 0; i < 3; i++) {
        glutSetWindow(subwindows[i]);
        int x0 = w*j/npanes, x1 = pane_enabled[i] ? w*(j+1)/npanes : x0;
        glutPositionWindow(x0, 0);
        glutReshapeWindow(x1-x0, h);
        if (pane_enabled[i])
            j = j + 1;
        glViewport(0, 0, x1-x0, h);
    }
}

void vertex (const Vec2 &x) {
    glVertex2d(x[0], x[1]);
}

void vertex (const Vec3 &x) {
    glVertex3d(x[0], x[1], x[2]);
}

void normal (const Vec3 &n) {
    glNormal3d(n[0], n[1], n[2]);
}

void color (const Vec3 &x) {
    glColor3d(x[0], x[1], x[2]);
}

// template <typename T> T lerp (double t, double a, double b, const T &x, const T &y) {return x + (t-a)/(b-a)*(y-x);}

Vec3 strain_color (const Face *face) {
    Mat3x2 F = derivative(face->v[0]->node->x, face->v[1]->node->x,
                          face->v[2]->node->x, face);
    Vec2 l = eigen_decomposition(F.t()*F).l;
    double s0 = sqrt(l[0]) - 1, s1 = sqrt(l[1]) - 1;
    double tens = clamp(100*s0, 0., 0.5), comp = clamp(-100*s1, 0., 0.5);
    return Vec3(1-tens, (1-tens)*(1-comp), (1-comp));
    // double s = max(abs(s0),abs(s1));
    // Vec3 c = s<0.1 ? Vec3(1,1,1) :
    //          s<0.2 ? lerp(s, 0.1, 0.2, Vec3(1,1,1), Vec3(.5,.5,1)) :
    //          s<0.5 ? lerp(s, 0.2, 0.5, Vec3(.5,.5,1), Vec3(.5,1,.5)) :
    //          s<1.0 ? lerp(s, 0.5, 1.0, Vec3(.5,1,.5), Vec3(1,.5,.5)) :
    //          Vec3(1,.5,.5);
    // return c;
}

Vec3 area_color (const Face *face) {
    Vec2 u0 = face->v[0]->u, u1 = face->v[1]->u, u2 = face->v[2]->u;
    double l0 = norm(u1-u2), l1 = norm(u2-u0), l2 = norm(u0-u1);
    double l = max(l0,l1,l2);
    double h = 2*face->a/l * 2/sqrt(3);
    double lmin = ::sim.cloths[0].remeshing.size_min/2,
           lmax = ::sim.cloths[0].remeshing.size_max/4;
    double a = clamp((log(l) - log(lmin))/(log(lmax) - log(lmin)), 0., 1.),
           b = clamp((log(h) - log(lmin))/(log(lmax) - log(lmin)), 0., 1.);
    // inspired by http://www.sron.nl/~pault/colourschemes.pdf Fig. 3
    Vec3 color = Vec3(0.8-0.6*b, 0.4+0.4*a-0.3*b, 0.5+0.2*b);
    return color*1.5 - 0.5*Vec3(0.5,0.5,0.5); // increase saturation by 50%
}

Vec3 normal_color (const Face *face) {
    return (face->n + Vec3(1,1,1))/2.;
}

Vec3 colormap (double x) {
    double t = abs(x);
    if (x < 0)
        return 0.9*Vec3(1, 1-t, 1-2*t);
    else
        return 0.9*Vec3(1-2*t, 1-t, 1);
}

Vec3 plasticity_color (const Face *face) {
    double s = norm_F(face->S_plastic)/1000;
    double d = face->damage;
    s = min(s/2, 0.5);
    d = min(d/2, 0.5);
    return Vec3(1-s, (1-s)*(1-d), (1-d));
    // return colormap(trace(face->S_plastic)/500);
}

void ambient_occlusion (const Node *node) {
    Vec3 H = Vec3(0);
    double a = 0;
    for (int v = 0; v < node->verts.size(); v++) {
        const Vert *vert = node->verts[v];
        for (int i = 0; i < vert->adje.size(); i++) {
            const Edge *edge = vert->adje[i];
            const Vert *vert1 = (edge->v[0]!=vert) ? edge->v[0] : edge->v[1];
            const Face *face0 = edge->adjf[0], *face1 = edge->adjf[1];
            if (!face0 || !face1)
                continue;
            Vec3 e = vert1->node->x - vert->node->x;
            H += (cross(e, face0->n) - cross(e, face1->n))/2.;
        }
        for (int f = 0; f < vert->adjf.size(); f++) {
            const Face *face = vert->adjf[f];
            int i = find(vert, face->v);
            const Vert *vert1 = face->v[NEXT(i)], *vert2 = face->v[PREV(i)];
            a += norm(cross(vert1->node->x - vert->node->x,
                            vert2->node->x - vert->node->x))/6.;
        }
    }
    double h = dot(H, node->n)/a;
    h /= 1000;
    // can't use min(h,0.) and max(h,0.) because of mysterious graphics card bug
    double af = (-1-1/(min(h,-1e-3)-1))*0.25, ab = (-1+1/(max(h,1e-3)+1))*0.25;
    float af4[4] = {af,af,af,1}, ab4[4] = {ab,ab,ab,1};
    glMaterialfv(GL_FRONT, GL_EMISSION, af4);
    glMaterialfv(GL_BACK, GL_EMISSION, ab4);
}

void draw_mesh_ms (const Mesh &mesh, bool set_color=false) {
    glBegin(GL_TRIANGLES);
    for (int i = 0; i < mesh.faces.size(); i++) {
        if (i % 64 == 0) {
            glEnd();
            glBegin(GL_TRIANGLES);
        }
        Face *face = mesh.faces[i];
        if (set_color)
            color(plasticity_color(face));
        for (int v = 0; v < 3; v++) {
            vertex(face->v[v]->u);
        }
    }
    glEnd();
}

void draw_meshes_ms (bool set_color=false) {
    for (int m = 0; m < meshes.size(); m++)
        draw_mesh_ms(*meshes[m], set_color);
}

void draw_ms_edges (const Mesh &mesh) {
    glLineWidth(2);
    glBegin(GL_LINES);
    for (int e = 0; e < mesh.edges.size(); e++) {
        if (e % 256 == 0) {
            glEnd();
            glBegin(GL_LINES);
        }
        Edge *edge = mesh.edges[e];
        color(colormap(edge->theta_ideal)*0.9);
        vertex(edge->v[0]->u);
        vertex(edge->v[1]->u);
    }
    glEnd();
}

void draw_ms_edges () {
    for (int m = 0; m < meshes.size(); m++)
        draw_ms_edges(*meshes[m]);
}

void draw_mesh_ps (const Mesh &mesh, bool set_color=false) {
    glBegin(GL_TRIANGLES);
    for (int i = 0; i < mesh.faces.size(); i++) {
        Face *face = mesh.faces[i];
        if (i % 256 == 0) {
            glEnd();
            glBegin(GL_TRIANGLES);
        }
        normal(normal<PS>(face));
        for (int v = 0; v < 3; v++)
            vertex(face->v[v]->node->y);
    }
    glEnd();
}

void draw_meshes_ps (bool set_color=false) {
    for (int m = 0; m < meshes.size(); m++)
        draw_mesh_ps(*meshes[m], set_color);
}

void draw_mesh_ws (const Mesh &mesh, bool set_color=false) {
    if (set_color)
        glDisable(GL_COLOR_MATERIAL);
    glBegin(GL_TRIANGLES);
    for (int i = 0; i < mesh.faces.size(); i++) {
        Face *face = mesh.faces[i];
        if (i % 256 == 0) {
            glEnd();
            glBegin(GL_TRIANGLES);
        }
        if (set_color) {
            // int c = find((Mesh*)&mesh, ::meshes);
            // static const float phi = (1+sqrt(5))/2;
            // double hue = c*(2 - phi)*2*M_PI; // golden angle
            // hue = -0.6*M_PI + hue; // looks better this way :/
            // if (face->label%2 == 1) hue += M_PI;
            // static Vec3 a = Vec3(0.92, -0.39, 0), b = Vec3(0.05, 0.12, -0.99);
            // Vec3 frt = Vec3(0.7,0.7,0.7) + (a*cos(hue) + b*sin(hue))*0.3,
            //      bak = frt*0.5 + Vec3(0.5,0.5,0.5);
            Vec3 frt = Vec3(1), bak = Vec3(1);
            float front[4] = {frt[0], frt[1], frt[2], 1},
                  back[4] = {bak[0], bak[1], bak[2], 1};
            glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, front);
            glMaterialfv(GL_BACK, GL_AMBIENT_AND_DIFFUSE, back);
            // color(area_color(face));
        }
        normal(face->n);
        for (int v = 0; v < 3; v++) {
            // ambient_occlusion(face->v[v]->node);
            vertex(face->v[v]->node->x);
        }
    }
    glEnd();
    if (set_color)
        glEnable(GL_COLOR_MATERIAL);
}

void draw_meshes_ws (bool set_color=false) {
    for (int m = 0; m < meshes.size(); m++)
        draw_mesh_ws(*meshes[m], set_color);
}

void draw_node_vels () {
    double dt = 0.01;
    glBegin(GL_LINES);
    for (int m = 0; m < meshes.size(); m++) {
        const Mesh &mesh = *meshes[m];
        for (int n = 0; n < mesh.nodes.size(); n++) {
            const Node *node = mesh.nodes[n];
            glColor3d(0,0,1);
            vertex(node->x);
            vertex(node->x + dt*node->v);
            glColor3d(1,0,0);
            vertex(node->x);
            vertex(node->x - dt*node->v);
        }
    }
    for (int o = 0; o < sim.obstacles.size(); o++) {
        const Mesh &mesh = sim.obstacles[o].get_mesh();
        for (int n = 0; n < mesh.nodes.size(); n++) {
            const Node *node = mesh.nodes[n];
            glColor3d(0,0,1);
            vertex(node->x);
            vertex(node->x + dt*node->v);
            glColor3d(1,0,0);
            vertex(node->x);
            vertex(node->x - dt*node->v);
        }
    }
    glEnd();
}

void draw_vert_accels () {
    double dt2 = 1e-6;
    glBegin(GL_LINES);
    for (int m = 0; m < meshes.size(); m++) {
        const Mesh &mesh = *meshes[m];
        for (int n = 0; n < mesh.nodes.size(); n++) {
            const Node *node = mesh.nodes[n];
            glColor3d(0,0,1);
            vertex(node->x);
            vertex(node->x + dt2*node->acceleration);
            glColor3d(1,0,0);
            vertex(node->x);
            vertex(node->x - dt2*node->acceleration);
        }
    }
    glEnd();
}

void draw_face_normals () {
    glColor3f(0,0,1);
    glBegin(GL_LINES);
    for (int m = 0; m < meshes.size(); m++) {
        const Mesh &mesh = *meshes[m];
        for (int f = 0; f < mesh.faces.size(); f++) {
            const Face *face = mesh.faces[f];
            double scale = 1e-1;
            Vec3 c = (face->v[0]->node->x + face->v[1]->node->x
                      + face->v[2]->node->x)/3.;
            Vec3 n = normal<WS>(face);
            vertex(c);
            vertex(c + scale*n);
        }
    }
    glEnd();
}

template <Space s> void draw_seam_or_boundary_edges () {
    glColor3f(0,0,0);
    glBegin(GL_LINES);
    for (int m = 0; m < meshes.size(); m++) {
        const Mesh &mesh = *meshes[m];
        for (int e = 0; e < mesh.edges.size(); e++) {
            const Edge *edge = mesh.edges[e];
            if (!is_seam_or_boundary(edge))
                continue;
            vertex(pos<s>(edge->v[0]->node));
            vertex(pos<s>(edge->v[1]->node));
        }
    }
    glEnd();
}

void directional_light (int i, const Vec3 &dir, Vec3 dif) {
    float diffuse[4] = {dif[0], dif[1], dif[2], 1};
    float position[4] = {dir[0], dir[1], dir[2], 0};
    glEnable(GL_LIGHT0+i);
    glLightfv(GL_LIGHT0+i, GL_DIFFUSE, diffuse);
    glLightfv(GL_LIGHT0+i, GL_POSITION, position);
}

void ambient_light (Vec3 a) {
    float ambient[4] = {a[0], a[1], a[2], 1};
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);
}

double aspect_ratio () {
    return (double)glutGet(GLUT_WINDOW_WIDTH)/glutGet(GLUT_WINDOW_HEIGHT);
}

void apply_view (const View &view) {
    glTranslatef(view.offset[0], view.offset[1], 0);
    glScalef(view.scale, view.scale, view.scale);
    glRotatef(view.lat, 1,0,0);
    glRotatef(view.lon, 0,1,0);
    // glTranslatef(view.offset[0], view.offset[1], view.offset[2]);
}

void display_material () {
    glClearColor(1,1,1,1);
    glClear(GL_COLOR_BUFFER_BIT);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    double a = aspect_ratio();
    gluOrtho2D(-a/2, a/2, -0.5, 0.5);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    apply_view(views[MaterialPane]);
    glColor3d(0.9,0.9,0.9);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    draw_meshes_ms(true);
    glColor4d(0,0,0, 0.2);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    draw_meshes_ms();
    // draw_ms_edges();
    glutSwapBuffers();
    GLenum errCode;
    const GLubyte *errString;
    if ((errCode = glGetError()) != GL_NO_ERROR) {
        errString = gluErrorString(errCode);
        cout << "OpenGL Error: " << errString << endl;
    }
}

void display_plastic () {
    // glClearColor(0,1,0,1);
    glClearColor(1,1,1,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(1,1);
    glEnable(GL_COLOR_MATERIAL);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45, aspect_ratio(), 0.1, 10);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(0, 0, -1);
    glEnable(GL_LIGHTING);
    glEnable(GL_NORMALIZE);
    directional_light(0, Vec3(1,1,1), Vec3(0.3,0.35,0.5));
    directional_light(1, Vec3(-1,0,0), Vec3(0.5,0.3,0.3));
    ambient_light(Vec3(0.5));
    apply_view(views[PlasticPane]);
    glColor3d(1,1,1);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    draw_meshes_ps();
    glColor4d(0,0,0,0.2);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    draw_meshes_ps();
    draw_seam_or_boundary_edges<PS>();
    glutSwapBuffers();
    GLenum errCode;
    const GLubyte *errString;
    if ((errCode = glGetError()) != GL_NO_ERROR) {
        errString = gluErrorString(errCode);
        cout << "OpenGL Error: " << errString << endl;
    }
}

#include "physics.hpp"
void draw_forces () {
}

void display_world () {
    glClearColor(1,1,1,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(1,1);
    glEnable(GL_COLOR_MATERIAL);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45, aspect_ratio(), 0.1, 10);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(0, 0, -1);
    glEnable(GL_LIGHTING);
    glEnable(GL_NORMALIZE);
    directional_light(0, Vec3(1,1,1), Vec3(0.3,0.35,0.5));
    directional_light(1, Vec3(-1,0,0), Vec3(0.5,0.3,0.3));
    ambient_light(Vec3(0.5));
    apply_view(views[WorldPane]);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    draw_meshes_ws(true);
    glColor3d(0.8,0.8,0.8);
    glEnable(GL_CULL_FACE);
    for(int i=0; i<sim.obstacles.size(); ++i)
        draw_mesh_ws(sim.obstacles[i].get_mesh());
    glDisable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);
    glColor4d(0,0,0,0.2);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    // draw_meshes_ws();
    draw_seam_or_boundary_edges<WS>();
    glColor3d(0,0,1);
    // draw_vert_vels();
    // draw_face_normals();
    // draw_vert_accels();
    // draw_forces();
    glutSwapBuffers();
    GLenum errCode;
    const GLubyte *errString;
    if ((errCode = glGetError()) != GL_NO_ERROR) {
        errString = gluErrorString(errCode);
        cout << "OpenGL Error: " << errString << endl;
    }
}

struct MouseState {
    bool down;
    int x, y;
    enum {ROTATE, TRANSLATE, SCALE} func;
} mouse_state;

void mouse (int button, int state, int x, int y) {
    mouse_state.down = (state == GLUT_DOWN);
    mouse_state.x = x;
    mouse_state.y = y;
    int pane = get_pane();
    if (pane == -1) {ECHO("i don't know what to do with this event"); return;}
    View &view = views[pane];
    if ((button == 3) || (button == 4)) { // It's a wheel event
        mouse_state.func = MouseState::SCALE;
        // Each wheel event reports like a button click, GLUT_DOWN then GLUT_UP
        if (state == GLUT_UP) return; // Disregard redundant GLUT_UP events
        if (button == 3) {
            view.scale *= 1.2;
        } else {
            view.scale /= 1.2;
        }
        glutPostRedisplay();
    } else if (button == GLUT_LEFT_BUTTON) {
        mouse_state.func = MouseState::ROTATE;
    } else if (button == GLUT_MIDDLE_BUTTON) {
        mouse_state.func = MouseState::TRANSLATE;
    }
}

void motion (int x, int y) {
    if (!mouse_state.down)
        return;
    int pane = get_pane();
    if (pane == -1) {ECHO("i don't know what to do with this event"); return;}
    View &view = views[pane];
    if (mouse_state.func == MouseState::ROTATE && pane != 0) {
        // don't rotate material space
        double speed = 0.25;
        view.lon += (x - mouse_state.x)*speed;
        view.lat += (y - mouse_state.y)*speed;
        view.lat = clamp(view.lat, -90., 90.);
    } else if (mouse_state.func == MouseState::TRANSLATE) {
        double speed = 1e-3;
        view.offset[0] += (x - mouse_state.x)*speed;
        view.offset[1] -= (y - mouse_state.y)*speed;
    }
    mouse_state.x = x;
    mouse_state.y = y;
    glutPostRedisplay();
}

void run_glut (const GlutCallbacks &cb) {
    int argc = 1;
    char argv0[] = "";
    char *argv = argv0;
    glutInit(&argc, &argv);
    glutInitDisplayMode(GLUT_RGBA|GLUT_DOUBLE|GLUT_DEPTH|GLUT_MULTISAMPLE);
    glutInitWindowSize(1280,720);
    int window = glutCreateWindow("Crumpling and Cracking");
    glutReshapeFunc(reshape);
    glutIdleFunc(cb.idle);
    double x[4] = {0, 1280/3, 1280*2/3, 1280};
    void (*display[3]) () = {display_material, display_plastic, display_world};
    for (int i = 0; i < 3; i++) {
        subwindows[i] = glutCreateSubWindow(window, x[i],0, x[i+1],720);
        glutDisplayFunc(display[i]);
        glutKeyboardFunc(cb.keyboard);
        glutSpecialFunc(cb.special);
        glutMouseFunc(mouse);
        glutMotionFunc(motion);
    }
    glutMainLoop();
}

void redisplay () {
    for (int i = 0; i < 3; i++) {
        glutSetWindow(subwindows[i]);
        glutPostRedisplay();
    }
}
