/*
  Copyright ©2013 The Regents of the University of California
  (Regents). All Rights Reserved. Permission to use, copy, modify, and
  distribute this software and its documentation for educational,
  research, and not-for-profit purposes, without fee and without a
  signed licensing agreement, is hereby granted, provided that the
  above copyright notice, this paragraph and the following two
  paragraphs appear in all copies, modifications, and
  distributions. Contact The Office of Technology Licensing, UC
  Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620,
  (510) 643-7201, for commercial licensing opportunities.

  IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
  DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGE.

  REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING
  DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
  IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/

#include "mesh.hpp"
#include "util.hpp"
#include <assert.h>
#include <cstdlib>
using namespace std;

template <typename T1, typename T2> void check (const T1 *p1, const T2 *p2,
                                                const vector<T2*> &v2) {
    if (p2 && find((T2*)p2, v2) == -1) {
        cout << p1 << "'s adjacent " << p2 << " is not accounted for" << endl;
        abort();
    }
}
template <typename T1, typename T2> void not_null (const T1 *p1, const T2 *p2) {
    if (!p2) {
        cout << "adjacent to " << p1 << " is null " << p2 << endl;
        abort();
    }
}
template <typename T1, typename T2> void not_any_null
    (const T1 *p1, T2 *const*p2, int n) {
    bool any_null = false;
    for (int i = 0; i < n; i++) if (!p2[i]) any_null = true;
    if (any_null) {
        cout << "adjacent to " << p1 << " one of these is null" << endl;
        for (int i = 0; i < n; i++) cout << p2[i] << endl;
        abort();
    }
}
template <typename T1, typename T2> void not_all_null
    (const T1 *p1, T2 *const*p2, int n) {
    bool all_null = true;
    for (int i = 0; i < n; i++) if (p2[i]) all_null = false;
    if (all_null) {
        cout << "adjacent to " << p1 << " all of these are null" << endl;
        for (int i = 0; i < n; i++) cout << p2[i] << endl;
        abort();
    }
}

bool check_that_pointers_are_sane (const Mesh &mesh) {
    for (int v = 0; v < mesh.verts.size(); v++) {
        const Vert *vert = mesh.verts[v];
        not_null(vert, vert->node);
        check(vert, vert->node, mesh.nodes);
        if (find((Vert*)vert, vert->node->verts) == -1) {
            cout << "vert " << vert << "'s node " << vert->node << " doesn't contain it" << endl;
            abort();
        }
        for (int i = 0; i < vert->adjv.size(); i++)
            check(vert, vert->adjv[i], mesh.verts);
        for (int i = 0; i < vert->adje.size(); i++)
            check(vert, vert->adje[i], mesh.edges);
        for (int i = 0; i < vert->adjf.size(); i++)
            check(vert, vert->adjf[i], mesh.faces);
    }
    for (int n = 0; n < mesh.nodes.size(); n++) {
        const Node *node = mesh.nodes[n];
        for (int i = 0; i < node->verts.size(); i++)
            check(node, node->verts[i], mesh.verts);
    }
    for (int e = 0; e < mesh.edges.size(); e++) {
        const Edge *edge = mesh.edges[e];
        for (int i = 0; i < 2; i++)
            check(edge, edge->v[i], mesh.verts);
        not_any_null(edge, edge->v, 2);
        for (int i = 0; i < 2; i++)
            check(edge, edge->adjv[i], mesh.verts);
        not_all_null(edge, edge->adjv, 2);
        for (int i = 0; i < 2; i++)
            check(edge, edge->adjf[i], mesh.faces);
        not_all_null(edge, edge->adjf, 2);
    }
    for (int f = 0; f < mesh.faces.size(); f++) {
        const Face *face = mesh.faces[f];
        for (int i = 0; i < 3; i++)
            check(face, face->v[i], mesh.verts);
        not_any_null(face, face->v, 3);
        for (int i = 0; i < 3; i++)
            check(face, face->adje[i], mesh.edges);
        not_any_null(face, face->adje, 3);
        for (int i = 0; i < 3; i++)
            check(face, face->adjf[i], mesh.faces);
        not_all_null(face, face->adjf, 3);
    }
    return true;
}

bool check_that_contents_are_sane (const Mesh &mesh) {
    // // TODO
    // for (int v = 0; v < mesh.verts.size(); v++) {
    //     const Vert *vert = mesh.verts[v];
    //     if (!isfinite(norm2(vert->x + vert->v + vert->n) + vert->a)) {
    //         cout << "Vertex " << name(vert) << " is " << vert->x << " "
    //              << vert->v << " " << vert->n << " " << vert->a << endl;
    //         return false;
    //     }
    // }
    // for (int f = 0; f < mesh.faces.size(); f++) {
    //     const Face *face = mesh.faces[f];
    //     if (!isfinite(norm2(face->n) + face->a)) {
    //         cout << "Face " << name(face) << " is " << face->n << " "
    //              << face->a << endl;
    //         return false;
    //     }
    // }
    return true;
}

// // Adjacency

// void compute_vert_adjacency (Vert* v1) {
//     v1->adjv.clear();
//     const vector<Face*> &adjfs = v1->adjf;
//     for (int i = 0; i < adjfs.size(); i++) {
//         const Face *face = adjfs[i];
//         for (int j = 0; j < 3; j++) {
//             Vert *v2 = face->v[j];
//             if (v2 == v1)
//                 continue;
//             include(v2, v1->adjv);
//             include(v1, v2->adjv); // needed for add_face update
//         }
//     }
// }

// void compute_face_adjacency (Face* f1) {
//     for (int i1 = 0; i1 < 3; i1++) {
//         if (f1->adjf[i1] != NULL)
//             continue;
//         Vert *v1 = f1->v[(i1+1)%3], *v2 = f1->v[(i1+2)%3];
//         const vector<Face*> &adjfs = v1->adjf;
//         for (int j = 0; j < adjfs.size(); j++) {
//             Face* f2 = adjfs[j];
//             if (f2 == f1)
//                 continue;
//             for (int i2 = 0; i2 < 3; i2++) {
//                 if (f2->v[(i2+1)%3] == v2 && f2->v[(i2+2)%3] == v1) {
//                     f1->adjf[i1] = f2;
//                     f2->adjf[i2] = f1;
//                 }
//             }
//         }
//     }
// }

// Material space data

void compute_face_ms_data (Face* face) {
    face->Dm = Mat2x2(face->v[1]->u - face->v[0]->u,
                      face->v[2]->u - face->v[0]->u);
    face->invDm = face->Dm.inv();
    face->a = det(face->Dm)/2;
    if (face->a == 0)
        face->invDm = Mat2x2(0);
}

void compute_edge_ms_data (Edge* edge) {
    edge->l = norm(edge->v[0]->u - edge->v[1]->u);
}

void compute_vert_ms_data (Vert* vert) {
    vert->a = 0;
    const vector<Face*> &adjfs = vert->adjf;
    for (int i = 0; i < adjfs.size(); i++) {
        Face const* face = adjfs[i];
        vert->a += face->a/3;
    }
}

void compute_node_ms_data (Node* node) {
    node->a = 0;
    for (int v = 0; v < node->verts.size(); v++)
        node->a += node->verts[v]->a;
}

void compute_ms_data (Mesh &mesh) {
    // compute material-space data for each face
    vector<Face*>::iterator face_it;
    for (face_it = mesh.faces.begin(); face_it != mesh.faces.end(); ++face_it)
        compute_face_ms_data(*face_it);
    // compute material-space data for each edge
    vector<Edge*>::iterator edge_it;
    for (edge_it = mesh.edges.begin(); edge_it != mesh.edges.end(); ++edge_it)
        compute_edge_ms_data(*edge_it);
    // compute material-space data for each vert
    vector<Vert*>::iterator vert_it;
    for (vert_it = mesh.verts.begin(); vert_it != mesh.verts.end(); ++vert_it)
        compute_vert_ms_data(*vert_it);
    for (int n = 0; n < mesh.nodes.size(); n++)
        compute_node_ms_data(mesh.nodes[n]);
    // now we need to recompute world-space data
    compute_ws_data(mesh);
}

// World-space data

void compute_face_ws_data (Face* face) {
    const Vec3 &x0 = face->v[0]->node->x,
               &x1 = face->v[1]->node->x,
               &x2 = face->v[2]->node->x;
    face->n = normalize(cross(x1-x0, x2-x0));
    // Mat3x2 F = derivative(x0, x1, x2, face);
    // SVD<3,2> svd = singular_value_decomposition(F);
    // Mat3x2 Vt_ = 0;
    // for (int i = 0; i < 2; i++)
    //     for (int j = 0; j < 2; j++)
    //         Vt_(i,j) = svd.Vt(i,j);
    // face->R = svd.U*Vt_;
    // face->F = svd.Vt.t()*diag(svd.s)*svd.Vt;
}

void compute_node_ws_data (Node* node) {
    node->n = Vec3(0);
    for (int v = 0; v < node->verts.size(); v++) {
        const Vert *vert = node->verts[v];
        const vector<Face*> &adjfs = vert->adjf;
        for (int i = 0; i < adjfs.size(); i++) {
            Face const* face = adjfs[i];
            int j = find(vert, face->v), j1 = (j+1)%3, j2 = (j+2)%3;
            Vec3 e1 = face->v[j1]->node->x - node->x,
                 e2 = face->v[j2]->node->x - node->x;
            node->n += cross(e1,e2)/(2*norm2(e1)*norm2(e2));
        }
    }
    node->n = normalize(node->n);
}

void compute_ws_data (Mesh &mesh) {
    // Shared iterators
    vector<Face*>::iterator face_it;
    // compute local frames for faces
    for (face_it = mesh.faces.begin(); face_it != mesh.faces.end(); ++face_it)
        compute_face_ws_data(*face_it);
    // compute local frames for verts
    for (int n = 0; n < mesh.nodes.size(); n++)
        compute_node_ws_data(mesh.nodes[n]);
}

// Mesh operations

Edge *get_edge (const Vert *v0, const Vert *v1) {
    for (int e = 0; e < v0->adje.size(); e++) {
        Edge *edge = v0->adje[e];
        if (edge->v[0] == v1 || edge->v[1] == v1)
            return edge;
    }
    return NULL;
}

Hinge get_hinge (const Edge *edge) {
    Hinge hinge;
    hinge.edge[0] = hinge.edge[1] = NULL;
    hinge.s[0] = hinge.s[1] = -1;
    const Node *node0 = edge->v[0]->node, *node1 = edge->v[1]->node;
    for (int v0 = 0; v0 < node0->verts.size(); v0++) {
        for (int v1 = 0; v1 < node1->verts.size(); v1++) {
            const Vert *vert0 = node0->verts[v0], *vert1 = node1->verts[v1];
            Edge *e = get_edge(vert0, vert1);
            if (!e)
                continue;
            if (e->adjf[0]) {
                int i = (e->v[0]->node==node0) ? 0 : 1;
                hinge.edge[i] = e;
                hinge.s[i] = 0;
            }
            if (e->adjf[1]) {
                int i = (e->v[0]->node==node0) ? 1 : 0;
                hinge.edge[i] = e;
                hinge.s[i] = 1;
            }
        }
    }
    return hinge;
}

void connect (Vert *vert, Node *node) {
    vert->node = node;
    include(vert, node->verts);
}

void Mesh::add (Vert *vert) {
    verts.push_back(vert);
    vert->node = NULL;
    vert->preserve = false;
    vert->adjv.clear();
    vert->adje.clear();
    vert->adjf.clear();
    vert->index = verts.size()-1;
}

void Mesh::remove (Vert* vert) {
    if (!vert->adje.empty() || !vert->adjf.empty()) {
        cout << "Error: can't delete vert " << vert
             << " as it still has " << vert->adje.size() << " edges and "
             << vert->adjf.size() << " faces attached to it." << endl;
        return;
    }
    exclude(vert, verts);
}

void Mesh::add (Node *node) {
    nodes.push_back(node);
    node->index = nodes.size()-1;
    for (int v = 0; v < node->verts.size(); v++)
        node->verts[v]->node = node;
}

void Mesh::remove (Node* node) {
    exclude(node, nodes);
}

void Mesh::add (Edge *edge) {
    edges.push_back(edge);
    edge->adjf[0] = edge->adjf[1] = NULL;
    edge->adjv[0] = edge->adjv[1] = NULL;
    edge->index = edges.size()-1;
    include(edge->v[0], edge->v[1]->adjv);
    include(edge->v[1], edge->v[0]->adjv);
    include(edge, edge->v[0]->adje);
    include(edge, edge->v[1]->adje);
}

void Mesh::remove (Edge *edge) {
    if (edge->adjf[0] || edge->adjf[1]) {
        cout << "Error: can't delete edge " << edge
             << " as it still has (a) face(s) attached to it." << endl;
        return;
    }
    exclude(edge, edges);
    exclude(edge->v[0], edge->v[1]->adjv);
    exclude(edge->v[1], edge->v[0]->adjv);
    exclude(edge, edge->v[0]->adje);
    exclude(edge, edge->v[1]->adje);
}

void add_edges_if_needed (Mesh &mesh, const Face *face) {
    for (int i = 0; i < 3; i++) {
        Vert *v0 = face->v[i], *v1 = face->v[NEXT(i)];
        if (get_edge(v0, v1) == NULL)
            mesh.add(new Edge(v0, v1));
    }
}

void Mesh::add (Face *face) {
    faces.push_back(face);
    face->index = faces.size()-1;
    // adjacency
    add_edges_if_needed(*this, face);
    for (int i = 0; i < 3; i++) {
        Vert *v0 = face->v[NEXT(i)], *v1 = face->v[PREV(i)];
        include(face, v0->adjf);
        Edge *e = get_edge(v0, v1);
        face->adje[i] = e;
        int side = e->v[0]==v0 ? 0 : 1;
        e->adjf[side] = face;
        e->adjv[side] = face->v[i];
        Face *aface = e->adjf[1-side];
        face->adjf[i] = aface;
        if (aface) {
            int j = NEXT(find(v0, aface->v));
            aface->adjf[j] = face;
        }
    }
}

void Mesh::remove (Face* face) {
    exclude(face, faces);
    // adjacency
    for (int i = 0; i < 3; i++) {
        Vert *v0 = face->v[NEXT(i)];
        exclude(face, v0->adjf);
        Edge *e = face->adje[i];
        int side = e->v[0]==v0 ? 0 : 1;
        e->adjf[side] = NULL;
        e->adjv[side] = NULL;
        Face *aface = face->adjf[i];
        if (aface) {
            int j = NEXT(find(v0, aface->v));
            aface->adjf[j] = NULL;
        }
    }
}

void update_indices (Mesh &mesh) {
    for (int v = 0; v < mesh.verts.size(); v++)
        mesh.verts[v]->index = v;
    for (int e = 0; e < mesh.edges.size(); e++)
        mesh.edges[e]->index = e;
    for (int f = 0; f < mesh.faces.size(); f++)
        mesh.faces[f]->index = f;
    for (int n = 0; n < mesh.nodes.size(); n++)
        mesh.nodes[n]->index = n;
}

void mark_vertices_to_preserve (Mesh &mesh) {
    for (int v = 0; v < mesh.verts.size(); v++) {
        Vert *vert = mesh.verts[v];
        if (is_seam_or_boundary(vert) || vert->label)
            vert->preserve = true;
    }
    for (int e = 0; e < mesh.edges.size(); e++) {
        Edge *edge = mesh.edges[e];
        if (edge->label)
            edge->v[0]->preserve = edge->v[1]->preserve = true;
    }
}

void apply_transformation_onto (const Mesh &start_state, Mesh &onto,
                                const Transformation &tr) {
    for (int n = 0; n < onto.nodes.size(); n++)
        onto.nodes[n]->x = tr.apply(start_state.nodes[n]->x);
    compute_ws_data(onto);
}

void apply_transformation (Mesh& mesh, const Transformation& tr) {
    apply_transformation_onto(mesh, mesh, tr);
}

// vector<Vec3> vert_positions (const Mesh &mesh) {
//     vector<Vec3> xs(mesh.verts.size());
//     for (int v = 0; v < mesh.verts.size(); v++)
//         xs[v] = mesh.verts[v]->x;
//     return xs;
// }

// vector<Vec3> vert_velocities (const Mesh &mesh) {
//     vector<Vec3> vs(mesh.verts.size());
//     for (int v = 0; v < mesh.verts.size(); v++)
//         vs[v] = mesh.verts[v]->v;
//     return vs;
// }

void update_x0 (Mesh &mesh) {
    for (int n = 0; n < mesh.nodes.size(); n++)
        mesh.nodes[n]->x0 = mesh.nodes[n]->x;
}

Mesh deep_copy (const Mesh &mesh0) {
    Mesh mesh1;
    for (int v = 0; v < mesh0.verts.size(); v++) {
        const Vert *vert0 = mesh0.verts[v];
        Vert *vert1 = new Vert(vert0->u, vert0->label);
        vert1->preserve = vert0->preserve;
        mesh1.add(vert1);
    }
    for (int n = 0; n < mesh0.nodes.size(); n++) {
        const Node *node0 = mesh0.nodes[n];
        Node *node1 = new Node(node0->y, node0->x, node0->v, node0->label);
        node1->verts.resize(node0->verts.size());
        for (int v = 0; v < node0->verts.size(); v++)
            node1->verts[v] = mesh1.verts[node0->verts[v]->index];
        mesh1.add(node1);
    }
    for (int e = 0; e < mesh0.edges.size(); e++) {
        const Edge *edge0 = mesh0.edges[e];
        Edge *edge1 = new Edge(mesh1.verts[edge0->v[0]->index],
                               mesh1.verts[edge0->v[1]->index],
                               edge0->label);
        mesh1.add(edge1);
    }
    for (int f = 0; f < mesh0.faces.size(); f++) {
        const Face *face0 = mesh0.faces[f];
        Face *face1 = new Face(mesh1.verts[face0->v[0]->index],
                               mesh1.verts[face0->v[1]->index],
                               mesh1.verts[face0->v[2]->index],
                               face0->label);
        mesh1.add(face1);
    }
    compute_ms_data(mesh1);
    return mesh1;
}

void delete_mesh (Mesh &mesh) {
    for (int v = 0; v < mesh.verts.size(); v++)
        delete mesh.verts[v];
    for (int n = 0; n < mesh.nodes.size(); n++)
        delete mesh.nodes[n];
    for (int e = 0; e < mesh.edges.size(); e++)
        delete mesh.edges[e];
    for (int f = 0; f < mesh.faces.size(); f++)
        delete mesh.faces[f];
    mesh.verts.clear();
    mesh.nodes.clear();
    mesh.edges.clear();
    mesh.faces.clear();
}
