/*
  Copyright ©2013 The Regents of the University of California
  (Regents). All Rights Reserved. Permission to use, copy, modify, and
  distribute this software and its documentation for educational,
  research, and not-for-profit purposes, without fee and without a
  signed licensing agreement, is hereby granted, provided that the
  above copyright notice, this paragraph and the following two
  paragraphs appear in all copies, modifications, and
  distributions. Contact The Office of Technology Licensing, UC
  Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620,
  (510) 643-7201, for commercial licensing opportunities.

  IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
  DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGE.

  REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING
  DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
  IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/

#include "runphysics.hpp"

#include "conf.hpp"
#include "io.hpp"
#include "misc.hpp"
#include "simulation.hpp"
#include "timer.hpp"
#include "util.hpp"
#include <boost/filesystem.hpp>
#include <cstdio>
#include <fstream>

using namespace std;

static string outprefix;
static fstream timingfile;

Simulation sim;
Timer fps;

static void save (const std::vector<Mesh*> &meshes, int frame);
static void save_obstacle_transforms (const std::vector<Obstacle> &obs, int frame, double time);
static void save_timings ();

void copy_file (const string &input, const string &output);

void init_physics (const string &json_file, string outprefix,
                   bool is_reloading) {
    load_json(json_file, sim);
    ::outprefix = outprefix;
    if (!outprefix.empty()) {
        ::timingfile.open(stringf("%stiming", outprefix.c_str()).c_str(),
                          is_reloading ? ios::out|ios::app : ios::out);
        // Make a copy of the config file for future use
        copy_file(json_file.c_str(), stringf("%sconf.json", outprefix.c_str()));
        // And copy over all the obstacles
        vector<Mesh*> base_meshes(sim.obstacles.size());
        for (int o = 0; o < sim.obstacles.size(); o++)
            base_meshes[o] = &sim.obstacles[o].base_mesh;
        save_objs(base_meshes, stringf("%sobs", outprefix.c_str()));
    }
    prepare(sim);
    if (!is_reloading) {
        relax_initial_state(sim);
    }
}

static void save (const vector<Mesh*> &meshes, int frame) {
    if (!outprefix.empty() && frame < 10000)
        save_objs(meshes, stringf("%s%04d", outprefix.c_str(), frame));
}

static void save_obstacle_transforms (const vector<Obstacle> &obs, int frame,
                                      double time) {
    if (!outprefix.empty() && frame < 10000) {
        for (int o = 0; o < obs.size(); o++) {
            Transformation trans = identity();
            if (obs[o].transform_spline)
                trans = get_dtrans(*obs[o].transform_spline, time).first;
            save_transformation(trans, stringf("%s%04dobs%02d.txt",
                                               outprefix.c_str(), frame, o));
        }
    }
}

static void save_timings () {
    static double old_totals[Simulation::nModules] = {};
    ostream &out = ::timingfile ? ::timingfile : cout;
    for (int i = 0; i < Simulation::nModules; i++) {
        out << sim.timers[i].total - old_totals[i] << " ";
        old_totals[i] = sim.timers[i].total;
    }
    out << endl;
}

void save (const Simulation &sim, int frame) {
    save(sim.cloth_meshes, frame);
    save_obstacle_transforms(sim.obstacles, frame, sim.time);
}

void sim_step() {
    fps.tick();
    step(sim);
    bool save_this_frame = (::frame % sim.save_frames == 0);
    int save_frame_num = ::frame / sim.save_frames;
    if(save_this_frame)
        save(sim, save_frame_num);
    save_timings();
    fps.tock();
    if (sim.time >= sim.duration)
        exit(EXIT_SUCCESS);
}

void offline_loop() {
    while (sim.time < sim.duration) {
        sim_step();
    }
}

void run_physics (const vector<string> &args) {
    assert(args.size() == 1 || args.size() == 2);
    string json_file = args[0];
    string outprefix = args.size()>1 ? args[1] : "";
    init_physics(json_file, outprefix, false);
    if (!outprefix.empty())
        save(sim, 0);
    offline_loop();
}

void init_resume(const vector<string> &args) {
    assert(args.size() == 2);
    string outprefix = args[0];
    string start_frame_str = args[1];
    // Get the initialization information
    ::frame = atoi(start_frame_str.c_str());
    // Load like we would normally begin physics
    init_physics(stringf("%sconf.json", outprefix.c_str()), outprefix, true);
    // Fix the necessary data
    ::frame *= sim.save_frames;
    sim.time = ::frame * sim.timestep;
    for(int i=0; i<sim.obstacles.size(); ++i) {
        sim.obstacles[i].get_mesh(sim.time);
    }
    int read_frame = ::frame/sim.save_frames;
    load_objs(sim.cloth_meshes, stringf("%s%04d",outprefix.c_str(),read_frame));
    prepare(sim);
}

void resume_physics (const vector<string> &args) {
    init_resume(args);
    offline_loop();
}

void copy_file (const string &input, const string &output) {
    if(input == output) {
        return;
    }
    if(boost::filesystem::exists(output)) {
        boost::filesystem::remove(output);
    }
    boost::filesystem::copy_file(
        input, output);
}
