/*
  Copyright ©2013 The Regents of the University of California
  (Regents). All Rights Reserved. Permission to use, copy, modify, and
  distribute this software and its documentation for educational,
  research, and not-for-profit purposes, without fee and without a
  signed licensing agreement, is hereby granted, provided that the
  above copyright notice, this paragraph and the following two
  paragraphs appear in all copies, modifications, and
  distributions. Contact The Office of Technology Licensing, UC
  Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620,
  (510) 643-7201, for commercial licensing opportunities.

  IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
  DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGE.

  REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING
  DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
  IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/

#include "simulation.hpp"

#include "collision.hpp"
#include "dynamicremesh.hpp"
#include "geometry.hpp"
#include "magic.hpp"
#include "nearobs.hpp"
#include "physics.hpp"
#include "plasticity.hpp"
#include "popfilter.hpp"
#include "proximity.hpp"
#include "separate.hpp"
#include "strainlimiting.hpp"
#include <iostream>
#include <fstream>
using namespace std;

static const bool verbose = false;
static const int proximity = Simulation::Proximity,
                 physics = Simulation::Physics,
                 strainlimiting = Simulation::StrainLimiting,
                 collision = Simulation::Collision,
                 remeshing = Simulation::Remeshing,
                 separation = Simulation::Separation,
                 popfilter = Simulation::PopFilter,
                 plasticity = Simulation::Plasticity;

void physics_step (Simulation &sim, const vector<Constraint*> &cons);
void plasticity_step (Simulation &sim, bool damage);
void strainlimiting_step (Simulation &sim, const vector<Constraint*> &cons);
void equilibration_step (Simulation &sim);
void collision_step (Simulation &sim);
void remeshing_step (Simulation &sim, bool initializing=false);

void validate_handles (const Simulation &sim);

void prepare (Simulation &sim) {
    sim.cloth_meshes.resize(sim.cloths.size());
    for (int c = 0; c < sim.cloths.size(); c++) {
        compute_masses(sim.cloths[c]);
        sim.cloth_meshes[c] = &sim.cloths[c].mesh;
        update_x0(*sim.cloth_meshes[c]);
    }
    sim.obstacle_meshes.resize(sim.obstacles.size());
    for (int o = 0; o < sim.obstacles.size(); o++) {
        sim.obstacle_meshes[o] = &sim.obstacles[o].get_mesh();
        update_x0(*sim.obstacle_meshes[o]);
    }
}

void relax_initial_state (Simulation &sim) {
    validate_handles(sim);
    remeshing_step(sim, true);
    // debug_save_meshes(sim.cloth_meshes, "initremeshing");
    remeshing_step(sim, true);
    // debug_save_meshes(sim.cloth_meshes, "initremeshing");
    plasticity_step(sim, false);
    if (::magic.equilibrate) {
        equilibration_step(sim);
        // debug_save_meshes(sim.cloth_meshes, "initequilib");
        remeshing_step(sim, true);
        // debug_save_meshes(sim.cloth_meshes, "initremeshing");
        equilibration_step(sim);
        // debug_save_meshes(sim.cloth_meshes, "initequilib");
    }
    plasticity_step(sim, false);
}

void validate_handles (const Simulation &sim) {
    for (int h = 0; h < sim.handles.size(); h++) {
        vector<Node*> nodes = sim.handles[h]->get_nodes();
        for (int n = 0; n < nodes.size(); n++) {
            if (!nodes[n]->verts[0]->preserve) {
                cout << "Constrained node " << nodes[n]->index << " will not be preserved by remeshing" << endl;
                abort();
            }
        }
    }
}

void update_obstacles (Simulation &sim); 
vector<Constraint*> get_constraints (Simulation &sim, bool include_proximity);
void delete_constraints (const vector<Constraint*> &cons);

int frame = 0;

void step (Simulation &sim) {
    ::magic.preserve_creases = false;
    sim.time += sim.timestep;
    update_obstacles(sim);
    vector<Constraint*> cons = get_constraints(sim, true);
    physics_step(sim, cons);
    // debug_save_meshes(sim.cloth_meshes, "physics");
    plasticity_step(sim, true);
    // debug_save_meshes(sim.cloth_meshes, "plasticity");
    strainlimiting_step(sim, cons);
    // debug_save_meshes(sim.cloth_meshes, "strain");
    collision_step(sim);
    // debug_save_meshes(sim.cloth_meshes, "collision");
    if ((::frame+1) % sim.remeshing_frames == 0)
        remeshing_step(sim);
    // debug_save_meshes(sim.cloth_meshes, "remeshing");
    delete_constraints(cons);
    ::frame++;
}

vector<Constraint*> get_constraints (Simulation &sim, bool include_proximity) {
    vector<Constraint*> cons;
    for (int h = 0; h < sim.handles.size(); h++)
        append(cons, sim.handles[h]->get_constraints(sim.time));
    if (include_proximity && sim.enabled[proximity]) {
        sim.timers[proximity].tick();
        append(cons, proximity_constraints(sim.cloth_meshes,
                                           sim.obstacle_meshes,
                                           sim.friction, sim.obs_friction));
        sim.timers[proximity].tock();
    }
    return cons;
}

void delete_constraints (const vector<Constraint*> &cons) {
    for (int c = 0; c < cons.size(); c++)
        delete cons[c];
}

// Steps

void update_velocities (vector<Mesh*> &meshes, vector<Vec3> &xold, double dt);
void project_constraints (Simulation &sim, const vector<Constraint*> &cons);

void rewind_mesh (Mesh &mesh, double dt);
void forward_mesh (Mesh &mesh, double dt);

void physics_step (Simulation &sim, const vector<Constraint*> &cons) {
    if (!sim.enabled[physics])
        return;
    sim.timers[physics].tick();
    for (int o = 0; o < sim.obstacle_meshes.size(); o++)
        rewind_mesh(*sim.obstacle_meshes[o], sim.timestep);
    for (int c = 0; c < sim.cloths.size(); c++) {
        implicit_update(sim.cloths[c], sim.gravity, sim.wind, cons,
                        sim.timestep, false);
    }
    for (int c = 0; c < sim.cloth_meshes.size(); c++)
        forward_mesh(*sim.cloth_meshes[c], sim.timestep);
    for (int o = 0; o < sim.obstacle_meshes.size(); o++)
        forward_mesh(*sim.obstacle_meshes[o], sim.timestep);
    sim.timers[physics].tock();
}

void rewind_mesh (Mesh &mesh, double dt) {
    for (int n = 0; n < mesh.nodes.size(); n++) {
        Node *node = mesh.nodes[n];
        node->v = (node->x - node->x0)/dt;
        node->x = node->x0;
    }
}

void forward_mesh (Mesh &mesh, double dt) {
    for (int n = 0; n < mesh.nodes.size(); n++)
        mesh.nodes[n]->x += mesh.nodes[n]->v*dt;
}

void plasticity_step (Simulation &sim, bool damage) {
    if (!sim.enabled[plasticity])
        return;
    sim.timers[plasticity].tick();
    for (int c = 0; c < sim.cloths.size(); c++) {
        double yield = 0;
        if (!damage)
            swap(yield, sim.cloths[c].material.yield_curv);
        plastic_update(sim.cloths[c]);
        optimize_plastic_embedding(sim.cloths[c]);
        if (!damage) {
            swap(yield, sim.cloths[c].material.yield_curv);
            Mesh &mesh = sim.cloths[c].mesh;
            for (int f = 0; f < mesh.faces.size(); f++)
                mesh.faces[f]->damage = 0;
            for (int e = 0; e < mesh.edges.size(); e++)
                mesh.edges[e]->damage = 0;
        }
    }
    sim.timers[plasticity].tock();
}

void strainlimiting_step (Simulation &sim, const vector<Constraint*> &cons) {
    if (!sim.enabled[strainlimiting])
        return;
    sim.timers[strainlimiting].tick();
    vector<Vec3> xold = node_positions(sim.cloth_meshes);
    strain_limiting(sim.cloth_meshes, get_strain_limits(sim.cloths), cons);
    update_velocities(sim.cloth_meshes, xold, sim.timestep);
    sim.timers[strainlimiting].tock();
}

void equilibration_step (Simulation &sim) {
    sim.timers[remeshing].tick();
    vector<Constraint*> cons; // ignore actual constraints
    double stiff = 1;
    swap(stiff, ::magic.handle_stiffness);
    for (int c = 0; c < sim.cloths.size(); c++) {
        Mesh &mesh = sim.cloths[c].mesh;
        for (int n = 0; n < mesh.nodes.size(); n++)
            mesh.nodes[n]->acceleration = Vec3(0);
        apply_pop_filter(sim.cloths[c], cons);
    }
    swap(stiff, ::magic.handle_stiffness);
    sim.timers[remeshing].tock();
    if (sim.enabled[collision]) {
        sim.timers[collision].tick();
        collision_response(sim.cloth_meshes, cons, sim.obstacle_meshes);
        sim.timers[collision].tock();
    }
}

void collision_step (Simulation &sim) {
    if (!sim.enabled[collision])
        return;
    sim.timers[collision].tick();
    vector<Vec3> xold = node_positions(sim.cloth_meshes);
    vector<Constraint*> cons = get_constraints(sim, false);
    collision_response(sim.cloth_meshes, cons, sim.obstacle_meshes);
    delete_constraints(cons);
    update_velocities(sim.cloth_meshes, xold, sim.timestep);
    sim.timers[collision].tock();
}

void remeshing_step (Simulation &sim, bool initializing) {
    if (!sim.enabled[remeshing])
        return;
    // initialize pop filter
    if (sim.enabled[popfilter] && !initializing) {
        sim.timers[popfilter].tick();
        vector<Constraint*> cons = get_constraints(sim, true);
        for (int c = 0; c < sim.cloths.size(); c++)
            init_pop_filter(sim.cloths[c], cons);
        delete_constraints(cons);
        sim.timers[popfilter].tock();
    }
    // copy old meshes
    vector<Mesh> old_meshes(sim.cloths.size());
    vector<Mesh*> old_meshes_p(sim.cloths.size()); // for symmetry in separate()
    for (int c = 0; c < sim.cloths.size(); c++) {
        old_meshes[c] = deep_copy(sim.cloths[c].mesh);
        old_meshes_p[c] = &old_meshes[c];
    }
    // back up residuals
    typedef vector<Residual> MeshResidual;
    vector<MeshResidual> res;
    if (sim.enabled[plasticity] && !initializing) {
        sim.timers[plasticity].tick();
        res.resize(sim.cloths.size());
        for (int c = 0; c < sim.cloths.size(); c++)
            res[c] = back_up_residuals(sim.cloths[c].mesh);
        sim.timers[plasticity].tock();
    }
    // remesh
    sim.timers[remeshing].tick();
    for (int c = 0; c < sim.cloths.size(); c++) {
        vector<Plane> planes = nearest_obstacle_planes(sim.cloths[c].mesh,
                                                       sim.obstacle_meshes);
        dynamic_remesh(sim.cloths[c], planes);
    }
    sim.timers[remeshing].tock();
    // restore residuals
    if (sim.enabled[plasticity] && !initializing) {
        sim.timers[plasticity].tick();
        for (int c = 0; c < sim.cloths.size(); c++)
            restore_residuals(sim.cloths[c].mesh, old_meshes[c], res[c]);
        sim.timers[plasticity].tock();
    }
    // separate
    if (sim.enabled[separation]) {
        sim.timers[separation].tick();
        separate(sim.cloth_meshes, old_meshes_p, sim.obstacle_meshes);
        sim.timers[separation].tock();
    }
    // apply pop filter
    if (sim.enabled[popfilter] && !initializing) {
        sim.timers[popfilter].tick();
        vector<Constraint*> cons = get_constraints(sim, true);
        for (int c = 0; c < sim.cloths.size(); c++)
            apply_pop_filter(sim.cloths[c], cons);
        delete_constraints(cons);
        sim.timers[popfilter].tock();
    }
    // delete old meshes
    for (int c = 0; c < sim.cloths.size(); c++)
        delete_mesh(old_meshes[c]);
}

void update_velocities (vector<Mesh*> &meshes, vector<Vec3> &xold, double dt) {
    double inv_dt = 1/dt;
#pragma omp parallel for
    for (int n = 0; n < xold.size(); n++) {
        Node *node = get_node(n, meshes);
        node->v += (node->x - xold[n])*inv_dt;
    }
}

void update_obstacles (Simulation &sim) {
    for (int o = 0; o < sim.obstacles.size(); o++)
        sim.obstacles[o].get_mesh(sim.time);
}

// Helper functions

int count_verts (const vector<Mesh*> &meshes) {
    int nv = 0;
    for (int m = 0; m < meshes.size(); m++) nv += meshes[m]->verts.size();
    return nv;
}

int count_nodes (const vector<Mesh*> &meshes) {
    int nn = 0;
    for (int m = 0; m < meshes.size(); m++) nn += meshes[m]->nodes.size();
    return nn;
}

int count_edges (const vector<Mesh*> &meshes) {
    int ne = 0;
    for (int m = 0; m < meshes.size(); m++) ne += meshes[m]->edges.size();
    return ne;
}

int count_faces (const vector<Mesh*> &meshes) {
    int nf = 0;
    for (int m = 0; m < meshes.size(); m++) nf += meshes[m]->faces.size();
    return nf;
}

int get_index (const Vert *vert, const vector<Mesh*> &meshes) {
    int v = 0;
    for (int m = 0; m < meshes.size(); m++) {
        const Mesh &mesh = *meshes[m];
        if (vert->index < mesh.verts.size() && vert == mesh.verts[vert->index])
            return v + vert->index;
        else v += mesh.verts.size();
    }
    return -1;
}

int get_index (const Node *node, const vector<Mesh*> &meshes) {
    int n = 0;
    for (int m = 0; m < meshes.size(); m++) {
        const Mesh &mesh = *meshes[m];
        if (node->index < mesh.nodes.size() && node == mesh.nodes[node->index])
            return n + node->index;
        else n += mesh.nodes.size();
    }
    return -1;
}

int get_index (const Edge *edge, const vector<Mesh*> &meshes) {
    int e = 0;
    for (int m = 0; m < meshes.size(); m++) {
        const Mesh &mesh = *meshes[m];
        if (edge->index < mesh.edges.size() && edge == mesh.edges[edge->index])
            return e + edge->index;
        else e += mesh.edges.size();
    }
    return -1;
}

int get_index (const Face *face, const vector<Mesh*> &meshes) {
    int f = 0;
    for (int m = 0; m < meshes.size(); m++) {
        const Mesh &mesh = *meshes[m];
        if (face->index < mesh.faces.size() && face == mesh.faces[face->index])
            return f + face->index;
        else f += mesh.faces.size();
    }
    return -1;
}

Vert *get_vert (int v, const vector<Mesh*> &meshes) {
    for (int m = 0; m < meshes.size(); m++)
        if (v < meshes[m]->verts.size())
            return meshes[m]->verts[v];
        else
            v -= meshes[m]->verts.size();
    return NULL;
}

Node *get_node (int n, const vector<Mesh*> &meshes) {
    for (int m = 0; m < meshes.size(); m++)
        if (n < meshes[m]->nodes.size())
            return meshes[m]->nodes[n];
        else
            n -= meshes[m]->nodes.size();
    return NULL;
}

Edge *get_edge (int e, const vector<Mesh*> &meshes) {
    for (int m = 0; m < meshes.size(); m++)
        if (e < meshes[m]->edges.size())
            return meshes[m]->edges[e];
        else
            e -= meshes[m]->edges.size();
    return NULL;
}

Face *get_face (int f, const vector<Mesh*> &meshes) {
    for (int m = 0; m < meshes.size(); m++)
        if (f < meshes[m]->faces.size())
            return meshes[m]->faces[f];
        else
            f -= meshes[m]->faces.size();
    return NULL;
}

vector<Vec3> node_positions (const vector<Mesh*> &meshes) {
    vector<Vec3> xs(count_nodes(meshes));
    for (int n = 0; n < xs.size(); n++)
        xs[n] = get_node(n, meshes)->x;
    return xs;
}
